from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class ChartEvents(HighchartsMeta):
    """Event listeners for the chart."""

    def __init__(self, **kwargs):
        self._add_series = None
        self._after_print = None
        self._before_print = None
        self._click = None
        self._drilldown = None
        self._drillup = None
        self._drillupall = None
        self._export_data = None
        self._fullscreen_close = None
        self._fullscreen_open = None
        self._load = None
        self._redraw = None
        self._render = None
        self._selection = None

        for attribute in dir(self):
            if attribute.startswith('_') and not attribute.startswith('__'):
                non_private_name = attribute[1:]
                setattr(self, non_private_name, kwargs.get(non_private_name, None))

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'chart.events'

    @property
    def add_series(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when a series is added to the chart
        after load time, using the JavaScript ``.addSeries()`` method.

        One parameter, ``event``, is passed to the JavaScript function, containing common
        event information. Through ``event.options`` you can access the series options
        that were passed to the ``.addSeries()`` method.

        Returning ``false`` prevents the series from being added.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._add_series

    @add_series.setter
    @class_sensitive(CallbackFunction)
    def add_series(self, value):
        self._add_series = value

    @property
    def after_print(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires after a chart has been printed through
        the context menu or the ``Chart.print()`` JavaScript method.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._after_print

    @after_print.setter
    @class_sensitive(CallbackFunction)
    def after_print(self, value):
        self._after_print = value

    @property
    def before_print(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires before a chart is printed through
        the context menu or the ``Chart.print()`` JavaScript method.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._before_print

    @before_print.setter
    @class_sensitive(CallbackFunction)
    def before_print(self, value):
        self._before_print = value

    @property
    def click(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the user clicks on the plot
        background. One parameter, ``event``, is passed to the JavaScript function,
        containing common event information.

        .. hint::

          Information on the clicked spot can be found in your JavaScript function through
          ``event.xAxis`` and ``event.yAxis``, which are arrays containing the axes of
          each dimension and each axis' value at the clicked spot.

          The primary axes are ``event.xAxis[0]`` and ``event.yAxis[0]``.

          Remember the unit of a datetime axis is milliseconds since 1970-01-01 00:00:00.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._click

    @click.setter
    @class_sensitive(CallbackFunction)
    def click(self, value):
        self._click = value

    @property
    def drilldown(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when a drilldown point is clicked,
        before the new series is added.

        This callback function is also used for asynchronous drilldown, where the
        JavaScript ``seriesOptions`` are not added using the configuration options, but
        are instead loaded asynchronously.

        .. warning::

          When the user clicks a category label to triggle multiple series drilldowns,
          one ``drilldown`` event is triggered per point in the category.

        One parameter, ``event``, is passed to the JavaScript function,
        containing common event information plus certain specific properties:

          * ``category``: if a category label was clicked, this will contain its index
          * ``originalEvent``: the original browser event (typically a click) that
            triggered the drilldown
          * ``point``: the originating point
          * ``points``: if a category label was clicked, this will be an array that holds
            all points corresponding to the category
          * ``seriesOptions``: options for the new series

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._drilldown

    @drilldown.setter
    @class_sensitive(CallbackFunction)
    def drilldown(self, value):
        self._drilldown = value

    @property
    def drillup(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the user drills up from a
        drilldown series.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._drillup

    @drillup.setter
    @class_sensitive(CallbackFunction)
    def drillup(self, value):
        self._drillup = value

    @property
    def drillupall(self) -> Optional[CallbackFunction]:
        """In a chart with multiple drilldown series, this JavaScript callback function
        fires after all the series have been drilled up.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._drillupall

    @drillupall.setter
    @class_sensitive(CallbackFunction)
    def drillupall(self, value):
        self._drillupall = value

    @property
    def export_data(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires while exporting data. This enables
        the modification of data rows before they are processed into their final format.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._export_data

    @export_data.setter
    @class_sensitive(CallbackFunction)
    def export_data(self, value):
        self._export_data = value

    @property
    def fullscreen_close(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when a fullscreen view is closed,
        either through a context menu item, the ``Esc`` key, or the JavaScript
        ``Chart.fullscreen.close()`` method.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._fullscreen_close

    @fullscreen_close.setter
    @class_sensitive(CallbackFunction)
    def fullscreen_close(self, value):
        self._fullscreen_close = value

    @property
    def fullscreen_open(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when a fullscreen view is opened,
        either through a context menu item or the JavaScript ``Chart.fullscreen.open()``
        method.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._fullscreen_open

    @fullscreen_open.setter
    @class_sensitive(CallbackFunction)
    def fullscreen_open(self, value):
        self._fullscreen_open = value

    @property
    def load(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the chart has finished loading
        (including images, for example in point markers). One parameter, ``event``, is
        passed to the JavaScript function, containing common event information.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._load

    @load.setter
    @class_sensitive(CallbackFunction)
    def load(self, value):
        self._load = value

    @property
    def redraw(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the chart is redrawn, either after
        a JavaScript call to ``chart.redraw()`` or after an axis, series, or point is
        modified in JavaScript with the ``redraw`` option set to ``true``.

        One parameter, ``event``, is passed to the JavaScript function, containing common
        event information.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._redraw

    @redraw.setter
    @class_sensitive(CallbackFunction)
    def redraw(self, value):
        self._redraw = value

    @property
    def selection(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when an area of the chart has been
        selected (selection is enabled by setting :class:`Chart.zoom_type`).

        One parameter, ``event``, is passed to the JavaScript function, containing common
        event information.

        .. hint::

          The default action for the selection event is to zoom the chart to the selected
          area, however this can be prevented by calling ``event.preventDefault()`` or
          returning ``false`` from your JavaScript callback function.

        .. hint::

          Information on the selected area can be found in your JavaScript function
          through ``event.xAxis`` and ``event.yAxis``, which are arrays containing the
          axes of each dimension and each axis' minimum and maximum value within the
          selected area.

          The primary axes are ``event.xAxis[0]`` and ``event.yAxis[0]``.

          Remember the unit of a datetime axis is milliseconds since 1970-01-01 00:00:00.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._selection

    @selection.setter
    @class_sensitive(CallbackFunction)
    def selection(self, value):
        self._selection = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'add_series': as_dict.get('addSeries', None),
            'after_print': as_dict.get('afterPrint', None),
            'before_print': as_dict.get('beforePrint', None),
            'click': as_dict.get('click', None),
            'drilldown': as_dict.get('drilldown', None),
            'drillup': as_dict.get('drillup', None),
            'drillupall': as_dict.get('drillupall', None),
            'export_data': as_dict.get('exportData', None),
            'fullscreen_close': as_dict.get('fullscreenClose', None),
            'fullscreen_open': as_dict.get('fullscreenOpen', None),
            'load': as_dict.get('load', None),
            'redraw': as_dict.get('redraw', None),
            'selection': as_dict.get('selection', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'addSeries': self.add_series,
            'afterPrint': self.after_print,
            'beforePrint': self.before_print,
            'click': self.click,
            'drilldown': self.drilldown,
            'drillup': self.drillup,
            'drillupall': self.drillupall,
            'exportData': self.export_data,
            'fullscreenClose': self.fullscreen_close,
            'fullscreenOpen': self.fullscreen_open,
            'load': self.load,
            'redraw': self.redraw,
            'selection': self.selection
        }

        return untrimmed


class BreadcrumbEvents(HighchartsMeta):
    """Event listeners for breadcrumbs."""

    def __init__(self, **kwargs):
        self._click = None

        for attribute in dir(self):
            if attribute.startswith('_') and not attribute.startswith('__'):
                non_private_name = attribute[1:]
                setattr(self, non_private_name, kwargs.get(non_private_name, None))

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'breadcrumb.events'

    @property
    def click(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the user clicks on the plot
        background. One parameter, ``event``, is passed to the JavaScript function,
        containing common event information.

        .. hint::

          Information on the clicked spot can be found in your JavaScript function through
          ``event.xAxis`` and ``event.yAxis``, which are arrays containing the axes of
          each dimension and each axis' value at the clicked spot.

          The primary axes are ``event.xAxis[0]`` and ``event.yAxis[0]``.

          Remember the unit of a datetime axis is milliseconds since 1970-01-01 00:00:00.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._click

    @click.setter
    @class_sensitive(CallbackFunction)
    def click(self, value):
        self._click = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'click': as_dict.get('click', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'click': self.click
        }

        return untrimmed


class NavigationEvents(HighchartsMeta):
    """Event listeners for the chart."""

    def __init__(self, **kwargs):
        self._close_popup = None
        self._deselect_button = None
        self._select_button = None
        self._show_popup = None

        for attribute in dir(self):
            if attribute.startswith('_') and not attribute.startswith('__'):
                non_private_name = attribute[1:]
                setattr(self, non_private_name, kwargs.get(non_private_name, None))

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'navigation.events'

    @property
    def close_popup(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when a popup should be closed, for
        example when clicking on an annotation again.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._close_popup

    @close_popup.setter
    @class_sensitive(CallbackFunction)
    def close_popup(self, value):
        self._close_popup = value

    @property
    def deselect_button(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when button state should change, for
        example after adding an annotation.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._deselect_button

    @deselect_button.setter
    @class_sensitive(CallbackFunction)
    def deselect_button(self, value):
        self._deselect_button = value

    @property
    def select_button(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires on a button click.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._select_button

    @select_button.setter
    @class_sensitive(CallbackFunction)
    def select_button(self, value):
        self._select_button = value

    @property
    def show_popup(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when when selecting an annotation.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._show_popup

    @show_popup.setter
    @class_sensitive(CallbackFunction)
    def show_popup(self, value):
        self._show_popup = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'close_popup': as_dict.get('closePopup', None),
            'deselect_button': as_dict.get('deselectButton', None),
            'select_button': as_dict.get('selectButton', None),
            'show_popup': as_dict.get('showPopup', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'closePopup': self.close_popup,
            'deselectButton': self.deselect_button,
            'selectButton': self.select_button,
            'showPopup': self.show_popup
        }

        return untrimmed


class PointEvents(HighchartsMeta):
    """Event listeners for Points."""

    def __init__(self, **kwargs):
        self._click = None
        self._drag = None
        self._drag_start = None
        self._drop = None
        self._mouse_out = None
        self._mouse_over = None
        self._remove = None
        self._select = None
        self._unselect = None
        self._update = None

        self.click = kwargs.get('click', None)
        self.drag = kwargs.get('drag', None)
        self.drag_start = kwargs.get('drag_start', None)
        self.drop = kwargs.get('drop', None)
        self.mouse_out = kwargs.get('mouse_out', None)
        self.mouse_over = kwargs.get('mouse_over', None)
        self.remove = kwargs.get('remove', None)
        self.select = kwargs.get('select', None)
        self.unselect = kwargs.get('unselect', None)
        self.update = kwargs.get('update', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.series.point.events'

    @property
    def click(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when a point is clicked.

        One parameter, ``event``, is passed to the function, containing common event
        information.

        If the :meth:`Series.allow_point_select` option is ``True``, the default action
        for the point's click event is to toggle the point's select state. Returning
        ``False`` from the JavaScript event handler function cancels this action.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._click

    @click.setter
    @class_sensitive(CallbackFunction)
    def click(self, value):
        self._click = value

    @property
    def drag(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires while dragging a point.

        The mouse event is passed in as parameter. The original data can be accessed from
        ``e.origin``, and the new point values can be accessed from ``e.newPoints``. If
        there is only a single point being updated, it can be accessed from ``e.newPoint``
        for simplicity, and its ID can be accessed from ``e.newPointId``. The this context
        is the point being dragged. To stop the default drag action, return ``false``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._drag

    @drag.setter
    @class_sensitive(CallbackFunction)
    def drag(self, value):
        self._drag = value

    @property
    def drag_start(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when starting to drag a point.

        In JavaScript, the mouse event object is passed in as an argument. If a drag
        handle is used, ``e.updateProp`` is set to the data property being dragged.
        The ``this`` context is the point.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._drag_start

    @drag_start.setter
    @class_sensitive(CallbackFunction)
    def drag_start(self, value):
        self._drag_start = value

    @property
    def drop(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when a point is droped (when dragging ends).

        The mouse event is passed in as parameter. The original data can be accessed from
        ``e.origin``, and the new point values can be accessed from ``e.newPoints``. If
        there is only a single point being updated, it can be accessed from ``e.newPoint``
        for simplicity, and its ID can be accessed from ``e.newPointId``. The this context
        is the point being dragged.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._drop

    @drop.setter
    @class_sensitive(CallbackFunction)
    def drop(self, value):
        self._drop = value

    @property
    def mouse_out(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the mouse leaves the area close to the
        point.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._mouse_out

    @mouse_out.setter
    @class_sensitive(CallbackFunction)
    def mouse_out(self, value):
        self._mouse_out = value

    @property
    def mouse_over(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the mouse enters the area close to the
        point. One parameter, ``event``, is passed to the function, containing common
        event information.

        Returning ``false`` cancels the default behavior, which is to show a tooltip for
        the point.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._mouse_over

    @mouse_over.setter
    @class_sensitive(CallbackFunction)
    def mouse_over(self, value):
        self._mouse_over = value

    @property
    def remove(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the point is removed using the
        (JavaScript) ``.remove()`` method.

        One parameter, ``event``, is passed to the function. Returning ``false`` cancels
        the operation.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._remove

    @remove.setter
    @class_sensitive(CallbackFunction)
    def remove(self, value):
        self._remove = value

    @property
    def select(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the point is selected either
        programmatically or following a click on the point.

        One parameter, ``event``, is passed to the function. Returning ``false`` cancels
        the operation.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._select

    @select.setter
    @class_sensitive(CallbackFunction)
    def select(self, value):
        self._select = value

    @property
    def unselect(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when the point is unselected either
        programmatically or following a click on the point.

        One parameter, ``event``, is passed to the function. Returning ``false`` cancels
        the operation.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._unselect

    @unselect.setter
    @class_sensitive(CallbackFunction)
    def unselect(self, value):
        self._unselect = value

    @property
    def update(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when the point is updated programmatically
        through the (JavaScript) ``.update()`` method.

        One parameter, ``event``, is passed to the function. The new point options can be
        accessed through ``event.options``. Returning ``false`` cancels the operation.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._update

    @update.setter
    def update(self, value):
        self._update = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'click': as_dict.get('click', None),
            'drag': as_dict.get('drag', None),
            'drag_start': as_dict.get('dragStart', None),
            'drop': as_dict.get('drop', None),
            'mouse_out': as_dict.get('mouseOut', None),
            'mouse_over': as_dict.get('mouseOver', None),
            'remove': as_dict.get('remove', None),
            'select': as_dict.get('select', None),
            'unselect': as_dict.get('unselect', None),
            'update': as_dict.get('update', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'click': self.click,
            'drag': self.drag,
            'dragStart': self.drag_start,
            'drop': self.drop,
            'mouseOut': self.mouse_out,
            'mouseOver': self.mouse_over,
            'remove': self.remove,
            'select': self.select,
            'unselect': self.unselect,
            'update': self.update
        }

        return untrimmed


class SeriesEvents(HighchartsMeta):
    """Event listeners for Series."""

    def __init__(self, **kwargs):
        self._after_animate = None
        self._checkbox_click = None
        self._click = None
        self._hide = None
        self._legend_item_click = None
        self._mouse_out = None
        self._mouse_over = None
        self._show = None

        self.after_animate = kwargs.get('after_animate', None)
        self.checkbox_click = kwargs.get('checkbox_click', None)
        self.click = kwargs.get('click', None)
        self.hide = kwargs.get('hide', None)
        self.legend_item_click = kwargs.get('legend_item_click', None)
        self.mouse_out = kwargs.get('mouse_out', None)
        self.mouse_over = kwargs.get('mouse_over', None)
        self.show = kwargs.get('show', None)

    @property
    def after_animate(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires after the series has finished its initial
        animation, or if animation is disabled, immediately as the series is displayed.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._after_animate

    @after_animate.setter
    @class_sensitive(CallbackFunction)
    def after_animate(self, value):
        self._after_animate = value

    @property
    def checkbox_click(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when the checkbox next to the series' name in
        the legend is clicked.

        One parameter, ``event``, is passed to the (JavaScript) function. The state of the
        checkbox is found by ``event.checked``. The checked item is found by
        ``event.item``. Return ``false`` to prevent the default action which is to toggle
        the select state of the series.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._checkbox_click

    @checkbox_click.setter
    @class_sensitive(CallbackFunction)
    def checkbox_click(self, value):
        self._checkbox_click = value

    @property
    def click(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when when the series is clicked.

        One parameter, ``event``, is passed to the (JavaScript) function, containing
        common event information. Additionally, ``event.point`` holds a pointer to the
        nearest point on the graph.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._click

    @click.setter
    @class_sensitive(CallbackFunction)
    def click(self, value):
        self._click = value

    @property
    def hide(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when the series is hidden after chart generation
        time, either by clicking the legend item or by calling (in JavaScript)
        ``.hide()``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._hide

    @hide.setter
    @class_sensitive(CallbackFunction)
    def hide(self, value):
        self._hide = value

    @property
    def legend_item_click(self) -> Optional[CallbackFunction]:
        """JavaScript function that fires when the legend item belonging to the series is
        clicked.

        One parameter, ``event``, is passed to the (JavaScript) function. The default
        action is to toggle the visibility of the series. This can be prevented by
        returning ``false`` or calling (in JavaScript) ``event.preventDefault()``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._legend_item_click

    @legend_item_click.setter
    @class_sensitive(CallbackFunction)
    def legend_item_click(self, value):
        self._legend_item_click = value

    @property
    def mouse_out(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the mouse leaves the graph.

        One parameter, ``event``, is passed to the (JavaScript) function, containing
        common event information. If the ``stickyTracking`` option is ``true``,
        the ``mouse_out`` event doesn't happen before the mouse enters another graph or
        leaves the plot area.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._mouse_out

    @mouse_out.setter
    @class_sensitive(CallbackFunction)
    def mouse_out(self, value):
        self._mouse_out = value

    @property
    def mouse_over(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the mouse enters the graph.

        One parameter, ``event``, is passed to the function, containing common event
        information.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._mouse_over

    @mouse_over.setter
    @class_sensitive(CallbackFunction)
    def mouse_over(self, value):
        self._mouse_over = value

    @property
    def show(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the series is shown after chart generation
        time, either by clicking the legend item or by calling (in JavaScript)
        ``.show()``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._show

    @show.setter
    @class_sensitive(CallbackFunction)
    def show(self, value):
        self._show = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'after_animate': as_dict.get('afterAnimate', None),
            'checkbox_click': as_dict.get('checkboxClick', None),
            'click': as_dict.get('click', None),
            'hide': as_dict.get('hide', None),
            'legend_item_click': as_dict.get('legendItemClick', None),
            'mouse_out': as_dict.get('mouseOut', None),
            'mouse_over': as_dict.get('mouseOver', None),
            'show': as_dict.get('show', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'afterAnimate': self.after_animate,
            'checkboxClick': self.checkbox_click,
            'click': self.click,
            'hide': self.hide,
            'legendItemClick': self.legend_item_click,
            'mouseOut': self.mouse_out,
            'mouseOver': self.mouse_over,
            'show': self.show
        }

        return untrimmed


class SimulationEvents(SeriesEvents):
    """Event listeners for series that involve simulation / layout.
    
    .. versionadded:: Highcharts Core for Python v.1.1.0 / Highcharts Core (JS) v.11.0.0
    
    """
    
    def __init__(self, **kwargs):
        self._after_simulation = None
        
        self.after_simulation = kwargs.get('after_simulation', None)
        
        super().__init__(**kwargs)
        
    @property
    def after_simulation(self) -> Optional[CallbackFunction]:
        """Event which fires after the simulation is ended and the layout is stable.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._after_simulation

    @after_simulation.setter
    @class_sensitive(CallbackFunction)
    def after_simulation(self, value):
        self._after_simulation = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'after_animate': as_dict.get('afterAnimate', None),
            'checkbox_click': as_dict.get('checkboxClick', None),
            'click': as_dict.get('click', None),
            'hide': as_dict.get('hide', None),
            'legend_item_click': as_dict.get('legendItemClick', None),
            'mouse_out': as_dict.get('mouseOut', None),
            'mouse_over': as_dict.get('mouseOver', None),
            'show': as_dict.get('show', None),
            
            'after_simulation': as_dict.get('afterSimulation', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'afterSimulation': self.after_simulation,
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed
    

class ClusterEvents(HighchartsMeta):
    """General event handlers for marker clusters."""

    def __init__(self, **kwargs):
        self._drill_to_cluster = None

        self.drill_to_cluster = kwargs.get('drill_to_cluster', None)

    @property
    def drill_to_cluster(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the cluster point is clicked and
        :meth:`Cluster.drill_to_cluster` is ``True``.

        One parameter, ``event``, is passed to the function. The default action is to zoom
        to the cluster points range. This can be prevented by calling (in JavaScript)
        ``event.preventDefault()``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._drill_to_cluster

    @drill_to_cluster.setter
    @class_sensitive(CallbackFunction)
    def drill_to_cluster(self, value):
        self._drill_to_cluster = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        return {
            'drill_to_cluster': as_dict.get('drillToCluster', None)
        }

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'drillToCluster': self.drill_to_cluster
        }

        return untrimmed


class AxisEvents(HighchartsMeta):
    """Event listeners for axes."""

    def __init__(self, **kwargs):
        self._after_breaks = None
        self._after_set_extremes = None
        self._point_break = None
        self._point_in_break = None
        self._set_extremes = None

        self.after_breaks = kwargs.get('after_breaks', None)
        self.after_set_extremes = kwargs.get('after_set_extremes', None)
        self.point_break = kwargs.get('point_break', None)
        self.point_in_break = kwargs.get('point_in_break', None)
        self.set_extremes = kwargs.get('set_extremes', None)

    @property
    def after_breaks(self) -> Optional[CallbackFunction]:
        """A JavaScript event function fired after breaks have rendered. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._after_breaks

    @after_breaks.setter
    @class_sensitive(CallbackFunction)
    def after_breaks(self, value):
        self._after_breaks = value

    @property
    def after_set_extremes(self) -> Optional[CallbackFunction]:
        """As opposed to :meth:`AxisEvent.set_extremes`, this JavaScript event fires after
        the final min and max values are computed and corrected for
        :meth:`GenericAxis.min_range`. Defaults to :obj:`None <python:None>`.

        Fires when the minimum and maximum is set for the axis, either by calling the
        (JavaScript) ``.setExtremes()`` method or by selecting an area in the chart. One
        parameter, ``event``, is passed to the function, containing common event
        information.

        The new user-set minimum and maximum values can be found (in JavaScript) by
        ``event.min`` and ``event.max``. These reflect the axis minimum and maximum in
        axis values. The actual data extremes are found in ``event.dataMin`` and
        ``event.dataMax``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._after_set_extremes

    @after_set_extremes.setter
    @class_sensitive(CallbackFunction)
    def after_set_extremes(self, value):
        self._after_set_extremes = value

    @property
    def point_break(self) -> Optional[CallbackFunction]:
        """A JavaScript event fired when a break from this axis occurs on a point.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._point_break

    @point_break.setter
    @class_sensitive(CallbackFunction)
    def point_break(self, value):
        self._point_break = value

    @property
    def point_in_break(self) -> Optional[CallbackFunction]:
        """A JavaScript event fired when a point falls inside a break from this axis.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._point_in_break

    @point_in_break.setter
    @class_sensitive(CallbackFunction)
    def point_in_break(self, value):
        self._point_in_break = value

    @property
    def set_extremes(self) -> Optional[CallbackFunction]:
        """JavaScript function which fires when the minimum and maximum is set for the
        axis, either by calling the (JavaScript) ``.setExtremes()`` method or by selecting
        an area in the chart.

        One parameter, ``event``, is passed to the function, containing common event
        information.

        The new user-set minimum and maximum values can be found by ``event.min`` and
        ``event.max``. These reflect the axis minimum and maximum in data values. When an
        axis is zoomed all the way out from the "Reset zoom" button, ``event.min`` and
        ``event.max`` are ``null``, and the new extremes are set based on
        ``this.dataMin`` and ``this.dataMax``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._set_extremes

    @set_extremes.setter
    @class_sensitive(CallbackFunction)
    def set_extremes(self, value):
        self._set_extremes = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'after_breaks': as_dict.get('afterBreaks', None),
            'after_set_extremes': as_dict.get('afterSetExtremes', None),
            'point_break': as_dict.get('pointBreak', None),
            'point_in_break': as_dict.get('pointInBreak', None),
            'set_extremes': as_dict.get('setExtremes', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'afterBreaks': self.after_breaks,
            'afterSetExtremes': self.after_set_extremes,
            'pointBreak': self.point_break,
            'pointInBreak': self.point_in_break,
            'setExtremes': self.set_extremes
        }

        return untrimmed


class MouseEvents(HighchartsMeta):
    """Collection of basic / standard event handlers for mouse interactions."""

    def __init__(self, **kwargs):
        self._click = None
        self._mousemove = None
        self._mouseout = None
        self._mouseover = None

        self.click = kwargs.get('click', None)
        self.mousemove = kwargs.get('mousemove', None)
        self.mouseout = kwargs.get('mouseout', None)
        self.mouseover = kwargs.get('mouseover', None)

    @property
    def click(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the user clicks on the plot
        band. One parameter, ``event``, is passed to the JavaScript function,
        containing common event information. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._click

    @click.setter
    @class_sensitive(CallbackFunction)
    def click(self, value):
        self._click = value

    @property
    def mousemove(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires on the mousemove event within the plot
        band. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._mousemove

    @mousemove.setter
    @class_sensitive(CallbackFunction)
    def mousemove(self, value):
        self._mousemove = value

    @property
    def mouseout(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the mouse moves away from the
        corner of the plot band. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._mouseout

    @mouseout.setter
    @class_sensitive(CallbackFunction)
    def mouseout(self, value):
        self._mouseout = value

    @property
    def mouseover(self) -> Optional[CallbackFunction]:
        """JavaScript callback function that fires when the user moves their mouse over
        the plot band. One parameter, ``event``, is passed to the JavaScript function,
        containing common event information. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._mouseover

    @mouseover.setter
    @class_sensitive(CallbackFunction)
    def mouseover(self, value):
        self._mouseover = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'click': as_dict.get('click', None),
            'mousemove': as_dict.get('mousemove', None),
            'mouseout': as_dict.get('mouseout', None),
            'mouseover': as_dict.get('mouseover', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'click': self.click,
            'mousemove': self.mousemove,
            'mouseout': self.mouseout,
            'mouseover': self.mouseover
        }

        return untrimmed


class SonificationEvents(HighchartsMeta):
    """Event handlers for sonification."""
    
    def __init__(self, **kwargs):
        self._after_update = None
        self._before_play = None
        self._before_update = None
        self._on_boundary_hit = None
        self._on_end = None
        self._on_play = None
        self._on_series_end = None
        self._on_series_start = None
        self._on_stop = None
        
        self.after_update = kwargs.get('after_update', None)
        self.before_play = kwargs.get('before_play', None)
        self.before_update = kwargs.get('before_update', None)
        self.on_boundary_hit = kwargs.get('on_boundary_hit', None)
        self.on_end = kwargs.get('on_end', None)
        self.on_play = kwargs.get('on_play', None)
        self.on_series_end = kwargs.get('on_series_end', None)
        self.on_series_start = kwargs.get('on_series_start', None)
        self.on_stop = kwargs.get('on_stop', None)

    @property
    def after_update(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called *after* updating the
        sonification.
        
        A context object is passed to the function, with properties ``chart`` and ``timeline``.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._after_update
    
    @after_update.setter
    @class_sensitive(CallbackFunction)
    def after_update(self, value):
        self._after_update = value

    @property
    def before_play(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called immediately when playback is requested.
        
        A context object is passed to the function, with properties ``chart`` and ``timeline``.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._before_play
    
    @before_play.setter
    @class_sensitive(CallbackFunction)
    def before_play(self, value):
        self._before_play = value

    @property
    def before_update(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called *before* updating the
        sonification.
        
        A context object is passed to the function, with properties ``chart`` and ``timeline``.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._before_update
    
    @before_update.setter
    @class_sensitive(CallbackFunction)
    def before_update(self, value):
        self._before_update = value

    @property
    def on_boundary_hit(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called when attempting to play an adjacent point
        or series, and there is none found. By defualt, a percussive sound is played.
        
        A context object is passed to the function, with properties ``chart``, ``timeline``, and ``attemptedNext``. The
        ``attemptedNext`` property is a boolean value that is ``true`` if the boundary hit was from trying to play the 
        next series/point, and ``false`` if it was from trying to play the previous.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._on_boundary_hit
    
    @on_boundary_hit.setter
    @class_sensitive(CallbackFunction)
    def on_boundary_hit(self, value):
        self._on_boundary_hit = value

    @property
    def on_end(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called when playback is completed.
        
        A context object is passed to the function, with properties ``chart``, ``timeline``, and ``pointsPlayed`` where
        ``pointsPlayed`` is an array of ``Point`` objects referencing data points related to the audio events played.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._on_end
    
    @on_end.setter
    @class_sensitive(CallbackFunction)
    def on_end(self, value):
        self._on_end = value

    @property
    def on_play(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called on play.
        
        A context object is passed to the function, with properties ``chart`` and ``timeline``.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._on_play
    
    @on_play.setter
    @class_sensitive(CallbackFunction)
    def on_play(self, value):
        self._on_play = value

    @property
    def on_series_end(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called when finished playing a series.
        
        A context object is passed to the function, with properties ``series`` and ``timeline``.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._on_series_end
    
    @on_series_end.setter
    @class_sensitive(CallbackFunction)
    def on_series_end(self, value):
        self._on_series_end = value

    @property
    def on_series_start(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called when starting to play a series.
        
        A context object is passed to the function, with properties ``series`` and ``timeline``.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._on_series_start
    
    @on_series_start.setter
    @class_sensitive(CallbackFunction)
    def on_series_start(self, value):
        self._on_series_start = value

    @property
    def on_stop(self) -> Optional[CallbackFunction]:
        """Event (Javascript) :term:`callback function` that is called on pause, cancel, or if playback is
        completed.
        
        A context object is passed to the function, with properties ``chart``, ``timeline``, and ``pointsPlayed`` where
        ``pointsPlayed`` is an array of ``Point`` objects referencing data points related to the audio events played.
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._on_stop
    
    @on_stop.setter
    @class_sensitive(CallbackFunction)
    def on_stop(self, value):
        self._on_stop = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'after_update': as_dict.get('afterUpdate', None),
            'before_play': as_dict.get('beforePlay', None),
            'before_update': as_dict.get('beforeUpdate', None),
            'on_boundary_hit': as_dict.get('onBoundaryHit', None),
            'on_end': as_dict.get('onEnd', None),
            'on_play': as_dict.get('onPlay', None),
            'on_series_end': as_dict.get('onSeriesEnd', None),
            'on_series_start': as_dict.get('onSeriesStart', None),
            'on_stop': as_dict.get('onStop', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'afterUpdate': self.after_update,
            'beforePlay': self.before_play,
            'beforeUpdate': self.before_update,
            'onBoundaryHit': self.on_boundary_hit,
            'onEnd': self.on_end,
            'onPlay': self.on_play,
            'onSeriesEnd': self.on_series_end,
            'onSeriesStart': self.on_series_start,
            'onStop': self.on_stop,
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class PanelOptions(HighchartsMeta):
    """Configuration of a panel used in 3D charts."""

    def __init__(self, **kwargs):
        self._color = None
        self._size = None
        self._visible = None

        self.color = kwargs.get('color', None)
        self.size = kwargs.get('size', None)
        self.visible = kwargs.get('visible', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'chart.options3d.side'

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or gradient for the 3D panel. Defaults to
        ``'transparent'``.

        :returns: The backgorund color for the 3D panel
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def size(self) -> Optional[int | float | Decimal]:
        """The thickness of the panel. Defaults to ``1``.

        :rtype: numeric
        """
        return self._size

    @size.setter
    def size(self, value):
        self._size = validators.numeric(value,
                                        allow_empty = True,
                                        minimum = 0)

    @property
    def visible(self) -> Optional[bool | str]:
        """Indicates whether to display the panel in the frame. Defaults to ``'default'``.

        Accepts:

          * ``True`` to always display
          * ``False`` to not display
          * ``'auto'`` to display the panel when it is behind data
          * ``'default'`` to display the panel it is behind data based on the axis layout,
            ignoring the user's point of view

        :rtype: :class:`bool <python:bool>` or :class:`str <python:str>` or
          :obj:`None <python:None>`
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if value is None:
            self._visible = None
        elif isinstance(value, bool):
            self._visible = value
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['auto', 'default']:
                raise errors.HighchartsValueError(f'visible access either True, False, '
                                                  f'"auto", or "default". Received: '
                                                  f'{value}')
            self._visible = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'color': as_dict.get('color', None),
            'size': as_dict.get('size', None),
            'visible': as_dict.get('visible', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'color': self.color,
            'size': self.size,
            'visible': self.visible
        }


class Frame(HighchartsMeta):
    """Provides the option to draw a frame around the charts by defining a bottom, front,
    and back panel."""

    def __init__(self, **kwargs):
        self._back = None
        self._bottom = None
        self._front = None
        self._left = None
        self._right = None
        self._side = None
        self._size = None
        self._top = None
        self._visible = None

        self.back = kwargs.get('back', None)
        self.bottom = kwargs.get('bottom', None)
        self.front = kwargs.get('front', None)
        self.left = kwargs.get('left', None)
        self.right = kwargs.get('right', None)
        self.size = kwargs.get('size', None)
        self.top = kwargs.get('top', None)
        self.visible = kwargs.get('visible', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'chart.options3d.frame'

    @property
    def back(self) -> Optional[PanelOptions]:
        """The back side of the frame around a 3D chart.

        :rtype: :class:`PanelOptions`
        """
        return self._back

    @back.setter
    @class_sensitive(PanelOptions)
    def back(self, value):
        self._back = value

    @property
    def bottom(self) -> Optional[PanelOptions]:
        """The bottom side of the frame around a 3D chart.

        :rtype: :class:`PanelOptions`
        """
        return self._bottom

    @bottom.setter
    @class_sensitive(PanelOptions)
    def bottom(self, value):
        self._bottom = value

    @property
    def front(self) -> Optional[PanelOptions]:
        """The front side of the frame around a 3D chart.

        :rtype: :class:`PanelOptions`
        """
        return self._front

    @front.setter
    @class_sensitive(PanelOptions)
    def front(self, value):
        self._front = value

    @property
    def left(self) -> Optional[PanelOptions]:
        """The left side of the frame around a 3D chart.

        :rtype: :class:`PanelOptions`
        """
        return self._left

    @left.setter
    @class_sensitive(PanelOptions)
    def left(self, value):
        self._left = value

    @property
    def right(self) -> Optional[PanelOptions]:
        """The right side of the frame around a 3D chart.

        :rtype: :class:`PanelOptions`
        """
        return self._right

    @right.setter
    @class_sensitive(PanelOptions)
    def right(self, value):
        self._right = value

    @property
    def top(self) -> Optional[PanelOptions]:
        """The top side of the frame around a 3D chart.

        :rtype: :class:`PanelOptions`
        """
        return self._top

    @top.setter
    @class_sensitive(PanelOptions)
    def top(self, value):
        self._top = value

    @property
    def size(self) -> Optional[int | float | Decimal]:
        """The thickness of the frame. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._size

    @size.setter
    def size(self, value):
        self._size = validators.numeric(value,
                                        allow_empty = True,
                                        minimum = 0)

    @property
    def visible(self) -> Optional[bool | str]:
        """Indicates whether to display the frame. Defaults to ``'default'``.

        Accepts:

          * ``True`` to always display
          * ``False`` to not display
          * ``'auto'`` to display the frame when it is behind data
          * ``'default'`` to display the panel it is behind data based on the axis layout,
            ignoring the user's point of view

        :rtype: :class:`bool <python:bool>` or :class:`str <python:str>` or
          :obj:`None <python:None>`
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if value is None:
            self._visible = None
        elif isinstance(value, bool):
            self._visible = value
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['auto', 'default']:
                raise errors.HighchartsValueError(f'visible access either True, False, '
                                                  f'"auto", or "default". Received: '
                                                  f'{value}')

            self._visible = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'back': as_dict.get('back', None),
            'bottom': as_dict.get('bottom', None),
            'front': as_dict.get('front', None),
            'left': as_dict.get('left', None),
            'right': as_dict.get('right', None),
            'size': as_dict.get('size', None),
            'top': as_dict.get('top', None),
            'visible': as_dict.get('visible', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'back': self.back,
            'bottom': self.bottom,
            'front': self.front,
            'left': self.left,
            'right': self.right,
            'size': self.size,
            'top': self.top,
            'visible': self.visible
        }


class Options3D(HighchartsMeta):
    """Options to render charts in three dimensions.

    .. note::

      This feature requires the JavaScript ``highcharts-3d.js`` module, found in the
      download package or online at
      `code.highcharts.com/highcharts-3d.js <https://code.highcharts.com/highcharts-3d.js>`_.

    """

    def __init__(self, **kwargs):
        self._alpha = None
        self._axis_label_position = None
        self._beta = None
        self._depth = None
        self._enabled = None
        self._fit_to_plot = None
        self._frame = None
        self._view_distance = None

        self.alpha = kwargs.get('alpha', None)
        self.axis_label_position = kwargs.get('axis_label_position', None)
        self.beta = kwargs.get('beta', None)
        self.depth = kwargs.get('depth', None)
        self.enabled = kwargs.get('enabled', None)
        self.fit_to_plot = kwargs.get('fit_to_plot', None)
        self.frame = kwargs.get('frame', None)
        self.view_distance = kwargs.get('view_distance', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'chart.options3d'

    @property
    def alpha(self) -> Optional[int | float | Decimal]:
        """One of two rotation angles for the chart. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        self._alpha = validators.numeric(value, allow_empty = True)

    @property
    def axis_label_position(self) -> Optional[constants.EnforcedNullType | str]:
        """Set to ``'auto'`` to automatically move the labels to the best edge.

        Defaults to :class:`EnforcedNull <EnforcedNullType>`` which indicates
        a JavaScript value of ``null`` (as opposed to :obj:`None <python:None>`) which
        results in ``undefined`` when converted JavaScript.

        :rtype: :class:`str` or :class:`EnforcedNullType` or :obj:`None <python:None>`
        """
        return self._axis_label_position

    @axis_label_position.setter
    def axis_label_position(self, value):
        if not value:
            self._axis_label_position = None
        elif isinstance(value, constants.EnforcedNullType):
            self._axis_label_position = constants.EnforcedNull
        else:
            value = validators.string(value)
            value = value.lower()
            if value != 'auto':
                raise errors.HighchartsValueError(f'axis_label_position accepts an empty '
                                                  f'value or "auto". Received: {value}')

            self._axis_label_position = value

    @property
    def beta(self) -> Optional[int | float | Decimal]:
        """One of two rotation angles for the chart. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._beta

    @beta.setter
    def beta(self, value):
        self._beta = validators.numeric(value, allow_empty = True)

    @property
    def depth(self) -> Optional[int | float | Decimal]:
        """The total depth of the chart. Defaults to ``25``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._depth

    @depth.setter
    def depth(self, value):
        self._depth = validators.numeric(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, renders the chart using the 3D functionality. Defaults to
        ``False``.

        :returns: Flag enabling or disabling 3D rendering.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def fit_to_plot(self) -> Optional[bool]:
        """If ``True``, automatically adjusts the 3d box to the chart plot area. Defaults
        to ``True``.

        :returns: Flag controlling the auto-adjustment of the 3d box to the chart plot
          area.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._fit_to_plot

    @fit_to_plot.setter
    def fit_to_plot(self, value):
        if value is None:
            self._fit_to_plot = None
        else:
            self._fit_to_plot = bool(value)

    @property
    def frame(self) -> Optional[Frame]:
        """Settings to draw a frame around the 3d box.

        :rtype: :class:`Frame`
        """
        return self._frame

    @frame.setter
    @class_sensitive(Frame)
    def frame(self, value):
        self._frame = value

    @property
    def view_distance(self) -> Optional[int | float | Decimal]:
        """Defines the distance the viewer is standing in front of the chart. Defaults
        to ``25``.

        .. note::

          This setting is important to calculate the perspective effect in column and
          scatter charts. It is not used for 3D pie charts.

        :rtype: numeric
        """
        return self._view_distance

    @view_distance.setter
    def view_distance(self, value):
        self._view_distance = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'alpha': as_dict.get('alpha', None),
            'axis_label_position': as_dict.get('axisLabelPosition', None),
            'beta': as_dict.get('beta', None),
            'depth': as_dict.get('depth', None),
            'enabled': as_dict.get('enabled', None),
            'fit_to_plot': as_dict.get('fitToPlot', None),
            'frame': as_dict.get('frame', None),
            'view_distance': as_dict.get('viewDistance', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'alpha': self.alpha,
            'axisLabelPosition': self.axis_label_position,
            'beta': self.beta,
            'depth': self.depth,
            'enabled': self.enabled,
            'fitToPlot': self.fit_to_plot,
            'frame': self.frame,
            'viewDistance': self.view_distance
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class AnnounceNewData(HighchartsMeta):
    """Options for announcing new data to screen reader users.

    .. tip::

      Useful for dynamic data applications and drilldown.

      Keep in mind that frequent announcements will not be useful to users, as they
      won't have time to explore the new data. For these applications, consider making
      snapshots of the data accessible, and do the announcements in batches.

    :returns: Configuration for the announcement of new data to screen reader users.
    :rtype: :class:`AnnounceNewData` or :obj:`None <python:None>`
    """

    def __init__(self, **kwargs):
        self._announcement_formatter = None
        self._enabled = None
        self._interrupt_user = None
        self._minimum_announcement_interval = None

        self.announcement_formatter = kwargs.get('announcement_formatter', None)
        self.enabled = kwargs.get('enabled', None)
        self.interrupt_user = kwargs.get('interrupt_user', None)
        self.minimum_announcement_interval = kwargs.get('minimum_announcement_interval',
                                                        None)

    @property
    def announcement_formatter(self) -> Optional[CallbackFunction]:
        """Optional JavaScript formatter callback for the announcement.

        Expects a string containing JavaScript code. This code should be a JavaScript
        function that up to three arguments:

          #. The first argument is always an array of all series that received updates.
             If an announcement is already queued, the series that received updates for
             that announcement are also included in this array.
          #. The second argument is provided if ``chart.addSeries`` was called
             (in JavaScript), and there is a new series. In that case, this argument is a
             reference to the new series.
          #. The third argument, similarly, is provided if ``series.addPoint`` was called
             (in JavaScript), and there is a new point. In that case, this argument is a
             reference to the new point.

        The JavaScript function should return a string with the text to announce to the
        user. Return an empty string to not announce anything. Return ``false`` to use the
        default announcement format.

        :returns: The code of the JavaScript function ot use as the formatter callback for
          the new data announcement.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._announcement_formatter

    @announcement_formatter.setter
    @class_sensitive(CallbackFunction)
    def announcement_formatter(self, value):
        self._announcement_formatter = value

    @property
    def enabled(self) -> Optional[bool]:
        """Enable the announcement of new data to screen reader users.

        Defaults to ``False``.

        :returns: Flag indicating whether new data announcements are enabled for the
          chart.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def interrupt_user(self) -> Optional[bool]:
        """Choose whether or not the announcements should interrupt the screen reader.
        Defaults to ``False``.

        If not enabled, the user will be notified once idle.

        .. hint::

          It is recommended **not** to enable this setting unless there is a specific
          reason to do so.

        :returns: Flag indicating whether new data announcements should interrupt the
          screen reader.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._interrupt_user

    @interrupt_user.setter
    def interrupt_user(self, value):
        if value is None:
            self._interrupt_user = None
        else:
            self._interrupt_user = bool(value)

    @property
    def minimum_announcement_interval(self) -> Optional[int]:
        """Minimum interval between announcements in milliseconds. Defaults to ``5000``.

        .. warning::

          Non-integer values will be coerced to an :class:`int <python:int>`.

        If new data arrives before this amount of time has passed, it is queued for
        announcement.

        If another new data event happens while an announcement is queued, the queued
        announcement is dropped, and the latest announcement is queued instead. Set to 0
        to allow all announcements, but be warned that frequent announcements are
        disturbing to users.

        :returns: The minimum interval between announcements, expressed in milliseconds.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`

        :raises ValueError: if a negative value is supplied
        """
        return self._minimum_announcement_interval

    @minimum_announcement_interval.setter
    def minimum_announcement_interval(self, value: int | float | Decimal):
        self._minimum_announcement_interval = validators.integer(value,
                                                                 allow_empty = True,
                                                                 coerce_value = True,
                                                                 minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'announcement_formatter': as_dict.get('announcementFormatter', None),
            'enabled': as_dict.get('enabled', None),
            'interrupt_user': as_dict.get('interruptUser', None),
            'minimum_announcement_interval': as_dict.get('minimumAnnouncementInterval',
                                                            None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'announcementFormatter': self.announcement_formatter,
            'enabled': self.enabled,
            'interruptUser': self.interrupt_user,
            'minimumAnnouncementInterval': self.minimum_announcement_interval
        }

        return untrimmed



# -*- coding: utf-8 -*-

"""
***********************************
tests._fixtures
***********************************

Fixtures used by the SQLAthanor test suite.

"""
import os
import pathlib
from copy import deepcopy
from collections import UserDict

import pytest

from validator_collection import checkers, validators
from highcharts_core import constants


class State(object):
    """Class to hold incremental test state."""
    # pylint: disable=too-few-public-methods
    pass


@pytest.fixture
def state(request):
    """Return the :class:`State` object that holds incremental test state."""
    # pylint: disable=W0108
    return request.cached_setup(
        setup = lambda: State(),
        scope = "session"
    )


@pytest.fixture
def run_pyspark_tests(request):
    """Return the ``--pyspark`` command-line option."""
    value = request.config.getoption("--pyspark")
    value = value.lower()
    if value in ['false', False, 0, 'no', 'no']:
        return False
    else:
        return True


@pytest.fixture
def run_download_tests(request):
    """Return the ``--downloads`` command-line option."""
    value = request.config.getoption("--downloads")
    value = value.lower()
    if value in ['false', False, 0, 'no', 'no']:
        return False
    else:
        return True


@pytest.fixture
def input_files(request):
    """Return the ``--inputs`` command-line option."""
    return request.config.getoption("--inputs")

@pytest.fixture
def create_output_directory(request):
    """Return the ``--create-output-directory`` command-line option."""
    value = request.config.getoption("--create-output-directory")
    value = value.lower()
    if value in ['false', False, 0, 'no', 'no']:
        return False
    else:
        return True


def check_input_file(input_directory, input_value, create_directory = False):
    inputs = os.path.abspath(input_directory)
    if not os.path.exists(input_directory) and not create_directory:
        raise AssertionError('input directory (%s) does not exist' % inputs)
    elif not os.path.exists(input_directory) and create_directory:
        pathlib.Path(input_directory).mkdir(parents = True, exist_ok = True)
    elif not os.path.isdir(input_directory):
        raise AssertionError('input directory (%s) is not a directory' % inputs)

    try:
        input_file = os.path.join(input_directory, input_value)
    except (TypeError, AttributeError):
        input_file = None

    if input_file is not None:
        input_value = input_file

    return input_value


def to_camelCase(variable_name):
    if '_' not in variable_name:
        return variable_name

    if 'url' in variable_name:
        variable_name = variable_name.replace('url', 'URL')

    camel_case = ''
    previous_character = ''
    for character in variable_name:
        if character != '_' and previous_character != '_':
            camel_case += character
            previous_character = character
        elif character == '_':
            previous_character = character
        elif character != '_' and previous_character == '_':
            camel_case += character.upper()
            previous_character = character

    return camel_case


def to_js_dict(original):
    print(f'\n{original}\n')
    as_dict = {}
    for key in original:
        if 'use_html' in key:
            new_key = 'useHTML'
        elif key == 'allow_html':
            new_key = 'allowHTML'
        elif 'utc' in key:
            new_key = 'useUTC'
        elif '_csv' in key:
            new_key = 'downloadCSV'
        elif '_jpeg' in key:
            new_key = 'downloadJPEG'
        elif '_pdf' in key:
            new_key = 'downloadPDF'
        elif '_png' in key:
            new_key = 'downloadPNG'
        elif '_svg' in key:
            new_key = 'downloadSVG'
        elif '_xls' in key:
            new_key = 'downloadXLS'
        elif '_atr' in key:
            new_key = key.replace('_atr', 'ATR')
        elif key == 'drillup_text':
            new_key = 'drillUpText'
        elif key == 'drillup_button':
            new_key = 'drillUpButton'
        elif key == 'thousands_separator':
            new_key = 'thousandsSep'
        elif key == 'measure_xy':
            new_key = 'measureXY'
        elif key == 'use_gpu_translations':
            new_key = 'useGPUTranslations'
        elif key == 'label_rank':
            new_key = 'labelrank'
        else:
            new_key = to_camelCase(key)

        as_dict[new_key] = original[key]

    return as_dict


def does_kwarg_value_match_result(kwarg_value, result_value):
    """Validate whether the value of ``kwarg_value`` matches the value of ``result_value``.

    :returns: ``True`` if match, ``False`` if not
    """
    print(f'EVALUATING KWARG_VALUE:\n{kwarg_value}')
    print(f'  against\n    {result_value}')
    if isinstance(kwarg_value, (dict,
                                UserDict)) and not isinstance(result_value, (dict,
                                                                             UserDict)):
        print('- converting KWARG dict back to object INSTANCE')
        result_cls = result_value.__class__
        try:
            test_value = result_cls.from_dict(kwarg_value)
        except AttributeError:
            if not result_value and not kwarg_value:
                return True
            else:
                return False

        ('-- converted INSTANCE as JS literal:')
        print(test_value.to_js_literal())
        print('-- original result as JS literal:')
        print(result_value.to_js_literal())

        return test_value == result_value
    elif not isinstance(kwarg_value, (int, float)) and isinstance(result_value, (int,
                                                                                 float)):
        print('- converting KWARG value into numeric')
        test_value = validators.numeric(kwarg_value)
        return test_value == result_value
    elif isinstance(kwarg_value, (int, float)) and not isinstance(result_value, (int,
                                                                                 float)):
        print('- converting RESULT value into numeric')
        result_value = validators.numeric(result_value)
        return test_value == result_value
    elif isinstance(kwarg_value, (dict,
                                  UserDict)) and isinstance(result_value, (dict,
                                                                           UserDict)):
        print('- checking two dict objects')
        if len(kwarg_value) != len(result_value):
            print('-- len does not match')
            return False
        for key in kwarg_value:
            print(f'CHECKING: {key}')
            if key == 'patternOptions':
                matches = does_kwarg_value_match_result(kwarg_value.get(key),
                                                        result_value.get('pattern'))
            else:
                matches = does_kwarg_value_match_result(kwarg_value.get(key),
                                                        result_value.get(key))
            if not matches:
                print(f'-- dict key ({key}) does not match')
                return False

        return True
    elif checkers.is_iterable(kwarg_value):
        print('- evaluating a KWARG value that is iterable')
        if hasattr(result_value, 'from_setter'):
            print('- converting iterable to a HighchartsMeta from_setter()')
            updated_kwarg_value = result_value.__class__.from_array([kwarg_value])[0]
            matches = does_kwarg_value_match_result(updated_kwarg_value, result_value)
            return matches
        elif len(kwarg_value) != len(result_value):
            print('-- len does not match')
            return False
        counter = 0
        for item in kwarg_value:
            print('-- processing items in KWARG iterable')
            result_item = result_value[counter]
            print(f'-- evaluating:\n   {item}')
            print(f'-- against:\n   {result_item}')
            item_match = does_kwarg_value_match_result(item, result_item)
            if not item_match:
                print('-- DOES NOT MATCH')
                return False
            counter += 1
    elif kwarg_value is None and result_value is not None:
        print('- converting None to "null" to check')
        if result_value == 'null' or isinstance(result_value, constants.EnforcedNullType):
            return True
        else:
            return False
    elif isinstance(kwarg_value, str) and isinstance(result_value, dict):
        print('- checking function string against function dict')
        if 'body' in result_value and kwarg_value.startswith('function'):
            return True
        else:
            return False
    else:
        print('- falling back to simple equivalency check')
        return kwarg_value == result_value

    return True


def trim_expected(expected):
    """Remove keys from ``expected`` or its children that should not be evaluated."""
    new_dict = {}
    if not isinstance(expected, dict):
        return expected
    for key in expected:
        if expected[key] is None:
            continue
        elif isinstance(expected[key], dict):
            trimmed_value = trim_expected(expected[key])
            if trimmed_value:
                new_dict[key] = trimmed_value
        elif checkers.is_iterable(expected[key]):
            trimmed_value = []
            for item in expected[key]:
                trimmed_item = trim_expected(item)
                if trimmed_item:
                    trimmed_value.append(trimmed_item)

            if trimmed_value:
                new_dict[key] = trimmed_value
        else:
            new_dict[key] = expected[key]

    return new_dict


def compare_js_literals(original, new):
    if not isinstance(original, str):
        original = str(original)
    if not isinstance(new, str):
        new = str(new)

    counter = 0
    for char in original:
        min_index = max(0, counter - 20)
        max_index = min(counter + 20, len(original))

        if new[counter] != char:
            print(f'\nMISMATCH FOUND AT ORIGINAL CHARACTER: {counter}')
            print(f'-- ORIGINAL: {original[min_index:max_index]}')
            print(f'-- NEW: {new[min_index:max_index]}')
            break

        counter += 1


def Class__init__(cls, kwargs, error):
    kwargs_copy = deepcopy(kwargs)
    if not error:
        result = cls(**kwargs)
        print(kwargs_copy)
        assert result is not None
        assert isinstance(result, cls) is True
        if checkers.is_type(result, 'GenericTypeOptions'):
            kwargs_copy['type'] = result.type
            #kwargs['type'] = result.type
        for key in kwargs_copy:
            print(f'CHECKING: {key}')
            if kwargs.get(key) and isinstance(kwargs_copy[key],
                                              str) and kwargs[key].startswith('function'):
                continue
            if kwargs.get(key) and isinstance(kwargs_copy[key],
                                              str) and kwargs[key].startswith('class'):
                continue
            if key == 'type' and checkers.is_type(result, 'GenericTypeOptions'):
                assert does_kwarg_value_match_result(kwargs_copy[key],
                                                     getattr(result, 'type'))
            elif key == 'margin' and checkers.is_type(result, 'ChartOptions'):
                print(f'CHECKING: {key}, which gets split over other values')
                if checkers.is_iterable(kwargs_copy[key]):
                    print('is iterable')
                    assert does_kwarg_value_match_result(kwargs_copy[key][0],
                                                         getattr(result, 'margin_top'))
                    assert does_kwarg_value_match_result(kwargs_copy[key][1],
                                                         getattr(result, 'margin_right'))
                    assert does_kwarg_value_match_result(kwargs_copy[key][2],
                                                         getattr(result, 'margin_bottom'))
                    assert does_kwarg_value_match_result(kwargs_copy[key][3],
                                                         getattr(result, 'margin_left'))
                elif kwargs_copy[key]:
                    print('not iterable')
                    print('checking margin_top')
                    print(result.margin_top)
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         getattr(result, 'margin_top'))
                    print('checking margin_right')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         getattr(result, 'margin_right'))
                    print('checking margin_bottom')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         getattr(result, 'margin_bottom'))
                    print('checking margin_left')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         getattr(result, 'margin_left'))
            else:
                print('not margin')
                kwarg_value = kwargs_copy[key]
                result_value = getattr(result, key)
                print(f'KWARG VALUE:\n{kwarg_value}')
                print(f'RESULT VALUE:\n{result_value}')
                assert does_kwarg_value_match_result(kwargs_copy[key],
                                                     getattr(result, key)) is True
    else:
        with pytest.raises(error):
            result = cls(**kwargs)


def Class__to_untrimmed_dict(cls, kwargs, error):
    kwargs_copy = deepcopy(kwargs)
    if not error:
        instance = cls(**kwargs)
        result = instance._to_untrimmed_dict()
        assert result is not None
        assert isinstance(result, dict) is True
        if checkers.is_type(instance, 'GenericTypeOptions'):
            print('updating the type!')
            kwargs_copy['type'] = instance.type

        for key in kwargs_copy:
            kwarg_value = kwargs_copy[key]
            result_value = result.get(key)

            if kwargs.get(key) and isinstance(kwargs_copy[key],
                                              str) and kwargs[key].startswith('function'):
                continue
            if kwargs.get(key) and isinstance(kwargs_copy[key],
                                              str) and kwargs[key].startswith('class'):
                continue
            if '_' not in key and (key != 'margin' and not checkers.is_type(instance,
                                                                            'ChartOptions')):
                assert does_kwarg_value_match_result(kwargs_copy[key],
                                                     result.get(key)) is True
            else:
                if 'use_html' in key:
                    print(f'CHECKING: {key}')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         result.get('useHTML')) is True
                elif key == 'allow_html':
                    print(f'CHECKING: {key}')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         result.get('allowHTML')) is True
                elif 'utc' in key:
                    print(f'CHECKING: {key}')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         result.get('useUTC')) is True
                elif 'url' in key:
                    print(f'CHECKING: {key}')
                    updated_key = key.replace('url', 'URL')
                    matches = does_kwarg_value_match_result(
                        kwargs_copy[key],
                        result.get(to_camelCase(updated_key))
                    )
                    assert matches is True
                elif '_kd_tree' in key:
                    updated_key = key.replace('_kd_tree', 'KDTree')
                    matches = does_kwarg_value_match_result(
                        kwargs_copy[key],
                        result.get(to_camelCase(updated_key))
                    )
                    assert matches is True
                elif key == 'use_gpu_translations':
                    updated_key = 'useGPUTranslations'
                    print(f'using key: {updated_key}')
                    matches = does_kwarg_value_match_result(
                        kwargs_copy[key],
                        result.get(to_camelCase(updated_key))
                    )
                    assert matches is True
                elif '_csv' in key:
                    print(f'CHECKING: {key}')
                    new_key = 'downloadCSV'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif '_jpeg' in key:
                    print(f'CHECKING: {key}')
                    new_key = 'downloadJPEG'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif '_pdf' in key:
                    print(f'CHECKING: {key}')
                    new_key = 'downloadPDF'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif '_png' in key:
                    print(f'CHECKING: {key}')
                    new_key = 'downloadPNG'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif '_svg' in key:
                    print(f'CHECKING: {key}')
                    new_key = 'downloadSVG'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif '_xls' in key:
                    print(f'CHECKING: {key}')
                    new_key = 'downloadXLS'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif '_atr' in key:
                    print(f'CHECKING: {key}')
                    new_key = key.replace('_atr', 'ATR')
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif key == 'drillup_text':
                    print(f'CHECKING: {key}')
                    new_key = 'drillUpText'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif key == 'drillup_button':
                    print(f'CHECKING: {key}')
                    new_key = 'drillUpButton'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif key == 'thousands_separator':
                    print(f'CHECKING: {key}')
                    new_key = 'thousandsSep'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif key == 'measure_xy':
                    print(f'CHECKING: {key}')
                    new_key = 'measureXY'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                elif key == 'pattern_options':
                    print(f'CHECKING: {key}')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         result.get('pattern')) is True
                elif isinstance(instance, UserDict):
                    print(f'CHECKING: {key}')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         result.get(key)) is True
                elif key == 'margin' and checkers.is_type(instance, 'ChartOptions'):
                    print(f'CHECKING: {key}, which gets split over other values')
                    if checkers.is_iterable(kwargs_copy[key]):
                        assert does_kwarg_value_match_result(kwargs_copy[key][0],
                                                             result.get('marginTop'))
                        assert does_kwarg_value_match_result(kwargs_copy[key][1],
                                                             result.get('marginRight'))
                        assert does_kwarg_value_match_result(kwargs_copy[key][2],
                                                             result.get('marginBottom'))
                        assert does_kwarg_value_match_result(kwargs_copy[key][3],
                                                             result.get('marginLeft'))
                    elif kwargs_copy[key]:
                        print('EVALUATING single margin value to list')
                        assert does_kwarg_value_match_result([kwargs_copy[key],
                                                              kwargs_copy[key],
                                                              kwargs_copy[key],
                                                              kwargs_copy[key]],
                                                             result.get('margin'))
                        assert does_kwarg_value_match_result(kwargs_copy[key],
                                                             result.get('marginTop'))
                        assert does_kwarg_value_match_result(kwargs_copy[key],
                                                             result.get('marginRight'))
                        assert does_kwarg_value_match_result(kwargs_copy[key],
                                                             result.get('marginBottom'))
                        assert does_kwarg_value_match_result(kwargs_copy[key],
                                                             result.get('marginLeft'))
                elif key == 'spacing' and checkers.is_type(instance, 'ChartOptions'):
                    print(f'CHECKING: {key}, which gets split over other values')
                    if checkers.is_iterable(kwargs_copy[key]):
                        assert does_kwarg_value_match_result(kwargs_copy[key][0],
                                                             result.get('spacingTop'))
                        assert does_kwarg_value_match_result(kwargs_copy[key][1],
                                                             result.get('spacingRight'))
                        assert does_kwarg_value_match_result(kwargs_copy[key][2],
                                                             result.get('spacingBottom'))
                        assert does_kwarg_value_match_result(kwargs_copy[key][3],
                                                             result.get('spacingLeft'))
                    elif kwargs_copy[key]:
                        print('EVALUATING single spacing value to list')
                        assert does_kwarg_value_match_result([kwargs_copy[key],
                                                              kwargs_copy[key],
                                                              kwargs_copy[key],
                                                              kwargs_copy[key]],
                                                             result.get('spacing'))
                elif key == 'label_rank':
                    new_key = 'labelrank'
                    matches = does_kwarg_value_match_result(kwargs_copy[key],
                                                            result.get(new_key))
                    assert matches is True
                else:
                    print(f'CHECKING: {key}')
                    assert does_kwarg_value_match_result(kwargs_copy[key],
                                                         result.get(to_camelCase(key))) is True
    else:
        with pytest.raises(error):
            instance = cls(**kwargs)
            result = instance._to_untrimmed_dict()


def Class_from_dict(cls, kwargs, error):
    if kwargs:
        as_dict = to_js_dict(deepcopy(kwargs))
    else:
        as_dict = {}

    if not error:
        instance = cls.from_dict(as_dict)
        assert instance is not None
        assert isinstance(instance, cls) is True
        for key in kwargs:
            print(f'CHECKING: {key}')
            if isinstance(kwargs[key], str) and kwargs[key].startswith('function'):
                continue
            if isinstance(kwargs[key], str) and kwargs[key].startswith('class'):
                continue
            elif key == 'margin' and checkers.is_type(instance, 'ChartOptions'):
                print(f'CHECKING: {key}, which gets split over other values')
                if checkers.is_iterable(kwargs[key]):
                    print('is iterable')
                    assert does_kwarg_value_match_result(kwargs[key][0],
                                                         getattr(instance, 'margin_top'))
                    assert does_kwarg_value_match_result(kwargs[key][1],
                                                         getattr(instance, 'margin_right'))
                    assert does_kwarg_value_match_result(kwargs[key][2],
                                                         getattr(instance, 'margin_bottom'))
                    assert does_kwarg_value_match_result(kwargs[key][3],
                                                         getattr(instance, 'margin_left'))
                elif kwargs[key]:
                    print('not iterable')
                    print('checking margin_top')
                    assert does_kwarg_value_match_result(kwargs[key],
                                                         getattr(instance, 'margin_top'))
                    print('checking margin_right')
                    assert does_kwarg_value_match_result(kwargs[key],
                                                         getattr(instance, 'margin_right'))
                    print('checking margin_bottom')
                    assert does_kwarg_value_match_result(kwargs[key],
                                                         getattr(instance, 'margin_bottom'))
                    print('checking margin_left')
                    assert does_kwarg_value_match_result(kwargs[key],
                                                         getattr(instance, 'margin_left'))
            else:
                kwarg_value = kwargs[key]
                if key in ['pattern_options', 'patternOptions']:
                    result_value = getattr(instance, 'pattern_options')
                elif key == 'type':
                    result_value = getattr(instance, 'type')
                else:
                    result_value = getattr(instance, key)
                print(kwarg_value)
                print(result_value)
                assert does_kwarg_value_match_result(kwarg_value, result_value)
    else:
        with pytest.raises(error):
            instance = cls.from_dict(as_dict)


def Class_to_dict(cls, kwargs, error):
    untrimmed_expected = to_js_dict(deepcopy(kwargs))
    expected = trim_expected(untrimmed_expected)
    check_dicts = True
    for key in expected:
        if not checkers.is_type(expected[key], (str, int, float, bool, list, dict)):
            check_dicts = False
        elif isinstance(expected[key], str) and expected[key].startswith('function'):
            check_dicts = False
        elif isinstance(expected[key], str) and expected[key].startswith('class'):
            check_dicts = False

    if not error:
        instance = cls(**kwargs)
        if checkers.is_type(instance, 'GenericTypeOptions'):
            expected['type'] = instance.type

        result = instance.to_dict()
        assert result is not None
        assert isinstance(result, dict) is True
        print('EXPECTED:')
        print(expected)
        print('RESULT:')
        print(result)
        if checkers.is_type(instance, ('MarkerAttributeObject')):
            check_dicts = False
        if check_dicts:
            if 'contextButton' not in result and 'context_button' not in result:
                assert len(expected) == len(result)
            else:
                assert len(result) == len(expected) + 1
            for key in expected:
                print(f'CHECKING: {key}')
                if key == 'patternOptions':
                    print('running special check for patternOptions')
                    assert does_kwarg_value_match_result(expected[key],
                                                         result.get('pattern')) is True
                elif key in ['contextButton', 'context_button']:
                    continue
                else:
                    assert does_kwarg_value_match_result(expected[key],
                                                         result.get(key)) is True
    else:
        with pytest.raises(error):
            instance = cls(**kwargs)
            result = instance.to_dict()


def append_plot_options_type(cls, as_str):
    """Append a JavaScript Literal ``type`` value to ``as_str`` based on ``cls``.

    :returns: An updated ``as_str``.
    :rtype: :class:`str <python:str>`
    """
    class_name = cls.__name__
    class_parents = cls.mro()
    is_generic_type_options = False
    for parent in class_parents:
        if 'GenericTypeOptions' in parent.__name__:
            is_generic_type_options = True
            break
    if not is_generic_type_options:
        return as_str
    class_name = class_name.replace('TypeOptions', '')
    class_name = class_name.replace('Options', '')
    if class_name.endswith('Series') and class_name != 'Series':
        class_name = class_name.replace('Series', '')

    class_name = class_name.lower()
    str_to_append = f"""type: '{class_name}'"""

    if str_to_append not in as_str:
        as_str = as_str[:-2] + """,\n  """ + str_to_append + """}"""

    return as_str


def Class_from_js_literal(cls, input_files, filename, as_file, error):
    input_file = check_input_file(input_files, filename)

    with open(input_file, 'r') as file_:
        as_str = file_.read()

    if as_file:
        input_string = input_file
    else:
        input_string = as_str

    as_str = append_plot_options_type(cls, as_str)

    if not error:
        #print('-------------------')
        #print('ORIGINAL VALIDATION')
        parsed_original, original_str = cls._validate_js_literal(as_str, range = False)
        #print('-------------')
        #print('ORIGINAL CALL')
        #print(as_str)
        result = cls.from_js_literal(input_string)
        assert result is not None
        assert isinstance(result, cls) is True

        as_js_literal = result.to_js_literal()
        #print('-----------------')
        #print('RESULT VALIDATION')
        if 'pattern:' in as_js_literal:
            as_js_literal = as_js_literal.replace('pattern:', 'patternOptions:')

        #print(as_js_literal)
        parsed_output, output_str = cls._validate_js_literal(as_js_literal, range = False)
        try:
            assert str(parsed_output) == str(parsed_original)
        except AssertionError as error:
            print('\n')
            compare_js_literals(str(parsed_original), str(parsed_output))
            raise error
    else:
        with pytest.raises(error):
            result = cls.from_js_literal(input_string)


def Class_from_js_literal_with_expected(cls,
                                        input_files,
                                        filename,
                                        expected_filename,
                                        as_file,
                                        error):
    input_file = check_input_file(input_files, filename)
    expected_file = check_input_file(input_files, expected_filename)

    with open(input_file, 'r') as file_:
        as_str = file_.read()

    with open(expected_file, 'r') as file_:
        expected_as_str = file_.read()

    if as_file:
        input_string = input_file
        expected_string = expected_file
    else:
        input_string = as_str
        expected_string = expected_as_str

    as_str = append_plot_options_type(cls, as_str)

    if not error:
        # print('-------------------')
        # print('ORIGINAL VALIDATION')
        parsed_expected, parsed_expected_str = cls._validate_js_literal(expected_string,
                                                                        range = False)
        # print('-------------')
        # print('ORIGINAL CALL')
        # print(as_str)
        result = cls.from_js_literal(input_string)
        assert result is not None
        assert isinstance(result, cls) is True

        as_js_literal = result.to_js_literal()
        # print('-----------------')
        # print('RESULT VALIDATION')
        if 'pattern:' in as_js_literal:
            as_js_literal = as_js_literal.replace('pattern:', 'patternOptions:')

        # print(as_js_literal)
        parsed_output, output_str = cls._validate_js_literal(as_js_literal, range = False)
        try:
            assert str(parsed_output) == str(parsed_expected)
        except AssertionError as error:
            print('\n')
            compare_js_literals(str(parsed_expected), str(parsed_output))
            raise error
    else:
        with pytest.raises(error):
            result = cls.from_js_literal(input_string)



from typing import Optional

import esprima
from esprima.error_handler import Error as ParseError

from validator_collection import validators, checkers

from highcharts_core import errors
from highcharts_core.decorators import validate_types
from highcharts_core.options import HighchartsOptions


class SharedOptions(HighchartsOptions):
    """Python object which can be used to configure global settings that can apply to
    multiple Highcharts charts when serialized to JavaScript."""

    def to_js_literal(self,
                      filename = None,
                      encoding = 'utf-8') -> Optional[str]:
        """Return the object represented as a :class:`str <python:str>` containing the
        JavaScript object literal.

        :param filename: The name of a file to which the JavaScript object literal should
          be persisted. Defaults to :obj:`None <python:None>`
        :type filename: Path-like

        :param encoding: The character encoding to apply to the resulting object. Defaults
          to ``'utf-8'``.
        :type encoding: :class:`str <python:str>`

        .. note::

          Returns a JavaScript string which applies the Highcharts global options. The
          JavaScript returned by this method takes the (pseudo-code) form:

          .. code-block:: javascript

            Highcharts.setOptions({... configuration options ... });

        :rtype: :class:`str <python:str>`
        """
        prefix = 'Highcharts.setOptions('
        options_body = super().to_js_literal(encoding = encoding)

        as_str = prefix + options_body + ');'

        if validators.path(filename, allow_empty = True):
            with open(filename, 'w', encoding = encoding) as file_:
                file_.write(as_str)

        return as_str

    @classmethod
    def _validate_js_literal(cls,
                             as_str,
                             range = True,
                             _break_loop_on_failure = False):
        """Parse ``as_str`` as a valid JavaScript literal object.

        :param as_str: The string to parse as a JavaScript literal object.
        :type as_str: :class:`str <python:str>`

        :param range: If ``True``, includes location and range data for each node in the
          AST returned. Defaults to ``False``.
        :type range: :class:`bool <python:bool>`

        :param _break_loop_on_failure: If ``True``, will not loop if the method fails to
          parse/validate ``as_str``. Defaults to ``False``.
        :type _break_loop_on_failure: :class:`bool <python:bool>`

        :returns: The parsed AST representation of ``as_str`` and the updated string.
        :rtype: 2-member :class:`tuple <python:tuple>` of :class:`esprima.nodes.Script`
          and :class:`str <python:str>`
        """
        try:
            parsed = esprima.parseScript(as_str, loc = range, range = range)
        except ParseError:
            try:
                parsed = esprima.parseModule(as_str, loc = range, range = range)
            except ParseError:
                if not _break_loop_on_failure:
                    as_str = f"""var randomVariable = {as_str}"""
                    return cls._validate_js_literal(as_str,
                                                    range = range,
                                                    _break_loop_on_failure = True)
                else:
                    raise errors.HighchartsParseError('._validate_js_function() expects '
                                                      'a str containing a valid '
                                                      'JavaScript function. Could not '
                                                      'find a valid function.')

        return parsed, as_str

    @classmethod
    def from_options(cls, options):
        """Create a
        :class:`SharedOptions <highcharts_core.global_options.shared_options.SharedOptions>`
        instance from a :class:`HighchartsOptions <highcharts_core.options.HighchartsOptions>`
        object.

        :param options: A :class:`HighchartsOptions` object to use for the shared options
          instance.
        :type options:
          :class:`HighchartsOptions <highcharts_core.options.HighchartsOptions>` or
          coercable

        :returns: A
          :class:`SharedOptions <highcharts_core.global_options.shared_options.SharedOptions>`
          instance
        :rtype: :class:`SharedOptions <highcharts_core.global_options.shared_options.SharedOptions>`
        """
        if not options:
            return cls()

        options = validate_types(options, types = (HighchartsOptions))
        options_as_dict = options.to_dict()

        instance = SharedOptions.from_dict(options_as_dict)

        return instance



"""Tests for the Python decorators defined in ``highcharts.decorators``."""
import json

import pytest

from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core import errors


class TestClass:
    def __init__(self, **kwargs):
        self.prop = None

        self.prop = kwargs.get('prop', None)

    @classmethod
    def from_dict(cls, as_dict):
        return cls(**as_dict)

    @classmethod
    def from_json(cls, as_json):
        as_dict = json.loads(as_json)
        return cls.from_dict(as_dict)


class TestDecoratedClass:
    def __init__(self, **kwargs):
        self._prop = None
        self._list_prop = None

        self.prop = kwargs.get('prop', None)
        self.list_prop = kwargs.get('list_prop', None)

    @property
    def prop(self):
        return self._prop

    @prop.setter
    @class_sensitive(types = TestClass,
                     allow_dict = True,
                     allow_json = True,
                     allow_none = True,
                     force_iterable = False)
    def prop(self, value):
        self._prop = value

    @property
    def list_prop(self):
        return self._list_prop

    @list_prop.setter
    @class_sensitive(types = TestClass,
                     allow_dict = True,
                     allow_json = True,
                     allow_none = True,
                     force_iterable = True)
    def list_prop(self, value):
        self._list_prop = value

    @classmethod
    def from_dict(cls, as_dict):
        return cls(**as_dict)

    @classmethod
    def from_json(cls, as_json):
        as_dict = json.loads(as_json)
        return cls.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error, result_class', [
    # allow_dict and everything works
    ({
        'value': {
            'prop': 123
        },
        'types': TestClass,
        'allow_dict': True
    }, None, TestClass),
    ({
        'value': {
            'prop': 123
        },
        'types': [TestClass, str],
        'allow_dict': True
    }, None, TestClass),
    # not allow_dict and raises an error
    ({
        'value': {'prop': 123},
        'types': TestClass,
        'allow_dict': False
    }, errors.HighchartsError, TestClass),
    # allow_json and everything works
    ({
        'value': '{"prop": 123}',
        'types': TestClass,
        'allow_json': True
    }, None, TestClass),
    ({
        'value': bytes('{"prop": 123}', encoding = 'utf8'),
        'types': TestClass,
        'allow_json': True
    }, None, TestClass),
    # not allow_json and raises an error
    ({
        'value': '{"prop": 123}',
        'types': TestClass,
        'allow_json': False
    }, errors.HighchartsError, TestClass),
    ({
        'value': bytes('{"prop": 123}', encoding = 'utf8'),
        'types': TestClass,
        'allow_json': False
    }, errors.HighchartsError, TestClass),
])
def test_validate_types(kwargs, error, result_class):
    if not error:
        value = kwargs.get('value', {})
        if isinstance(value, dict):
            prop_value = value.get('prop', None)
        else:
            prop_value = None

        result = validate_types(**kwargs)
        assert isinstance(result, result_class) is True
        assert result.prop is not None
        if prop_value:
            assert result.prop == prop_value

    else:
        with pytest.raises(error):
            result = validate_types(**kwargs)


@pytest.mark.parametrize('kwargs, error', [
    # everything works
    ({
        'value': None,
        'types': TestClass,
        'allow_none': True
    }, None),
    ({
        'value': '',
        'types': TestClass,
        'allow_none': True
    }, None),
    # fails
    ({
        'value': None,
        'types': TestClass,
        'allow_none': False
    }, errors.HighchartsError),
    ({
        'value': '',
        'types': TestClass,
        'allow_none': False
    }, errors.HighchartsError),
])
def test_validate_types_none(kwargs, error):
    if not error:
        result = validate_types(**kwargs)
        assert result is None
    else:
        with pytest.raises(error):
            result = validate_types(**kwargs)


@pytest.mark.parametrize('kwargs, error, result_class', [
    # force_iterable and allow_dict and everything works
    ({
        'value': [{'prop': 123}, {'prop': 456}],
        'types': TestClass,
        'allow_dict': True,
        'force_iterable': True
    }, None, TestClass),
    # not allow iterable
    ({
        'value': [{'prop': 123}, {'prop': 456}],
        'types': TestClass,
        'allow_dict': True,
        'force_iterable': False
    }, errors.HighchartsError, TestClass),
    # force_iterable and allow_json and everything works
    ({
        'value': '[{"prop": 123}, {"prop": 456}]',
        'types': TestClass,
        'allow_dict': True,
        'force_iterable': True,
        'allow_json': True
    }, None, TestClass),
    # force_iterable and allow_json and everything works
    ({
        'value': bytes('[{"prop": 123}, {"prop": 456}]', encoding = 'utf8'),
        'types': TestClass,
        'allow_dict': True,
        'force_iterable': True,
        'allow_json': True
    }, None, TestClass),
    # force_iterable and allow_empty and everything works
    ({
        'value': [],
        'types': TestClass,
        'allow_dict': True,
        'force_iterable': True,
        'allow_json': True,
        'allow_none': True
    }, None, TestClass),

])
def test_validate_types_list(kwargs, error, result_class):
    if not error:
        value = kwargs.get('value', None)
        result = validate_types(**kwargs)
        assert isinstance(result, list) is True
        if kwargs.get('allow_dict', False) and isinstance(value, list):
            assert len(kwargs.get('value', [])) == len(result)
        for item in result:
            assert isinstance(item, result_class) is True
            assert item.prop is not None

    else:
        with pytest.raises(error):
            result = validate_types(**kwargs)


@pytest.mark.parametrize('value, error, result_class', [
    # dict
    ({ 'prop': 123 }, None, TestClass),
    # json
    ('{ "prop": 123 }', None, TestClass),
    # none
    (None, None, TestClass),
    # list and fails
    ([{ 'prop': 123 }, {'prop': 456 }], errors.HighchartsError, TestClass)
])
def test_class_sensitive(value, error, result_class):
    test_instance = TestDecoratedClass()
    if not error:
        test_instance.prop = value
        if value:
            assert test_instance.prop is not None
            assert isinstance(test_instance.prop, result_class)
        else:
            assert test_instance.prop is None
    else:
        with pytest.raises(error):
            test_instance.prop = value


@pytest.mark.parametrize('value, error, result_class', [
    # dict
    ([{ 'prop': 123 }, { 'prop': 456 }], None, TestClass),
    # json
    ('[{ "prop": 123 }, { "prop": 456 }]', None, TestClass),
    # none
    (None, None, TestClass),
    # empty list
    ([], None, TestClass)
])
def test_class_sensitive_list(value, error, result_class):
    test_instance = TestDecoratedClass()
    if not error:
        test_instance.list_prop = value
        if value:
            assert test_instance.list_prop is not None
            assert isinstance(test_instance.list_prop, list)
        elif value is not None:
            assert len(test_instance.list_prop) == 0
        else:
            assert test_instance.list_prop is None
    else:
        with pytest.raises(error):
            test_instance.list_prop = value



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.zones import Zone as cls
from highcharts_core.utility_classes.zones import ClusterZone as cls2
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'class_name': 'some-class-name1',
      'color': '#999999',
      'dash_style': 'Solid',
      'fill_color': '#cccccc',
      'value': 123
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_Zone__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_Zone__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_Zone_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_Zone_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/zones/01.js', False, None),

    ('utility_classes/zones/error-01.js', False, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),

    ('utility_classes/zones/01.js', True, None),

    ('utility_classes/zones/error-01.js', True, (errors.HighchartsValueError,
                                                 errors.HighchartsParseError,
                                                 JSONDecodeError,
                                                 TypeError,
                                                 ValueError)),
])
def test_Zone_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


STANDARD_PARAMS_2 = [
    ({}, None),
    ({
     'class_name': 'classname-1',
     'from_': 1,
     'marker': {
       'fillColor': '#cccccc',
       'enabled': True,
       'height': 24,
       'lineWidth': 2,
       'radius': 2,
       'states': {
           'hover': {
               'enabled': True
           }
       },
       'symbol': 'circle',
       'width': 48
     },
     'to': 123
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_ClusterZone__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_ClusterZone__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_ClusterZone_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_ClusterZone_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/zones/02.js', False, None),

    ('utility_classes/zones/error-03.js', False, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),

    ('utility_classes/zones/02.js', True, None),

    ('utility_classes/zones/error-03.js', True, (errors.HighchartsValueError,
                                                 errors.HighchartsParseError,
                                                 JSONDecodeError,
                                                 TypeError,
                                                 ValueError)),
])
def test_ClusterZone_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)



from typing import Optional

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.metaclasses import HighchartsMeta


class SeriesNavigation(HighchartsMeta):
    """Options for the keyboard navigation of data points and series."""

    def __init__(self, **kwargs):
        self._mode = None
        self._point_navigation_enabled_threshold = None
        self._remember_point_focus = None
        self._skip_null_points = None

        self.mode = kwargs.get('mode', None)
        self.point_navigation_enabled_threshold = kwargs.get('point_navigation_enabled_threshold',
                                                             None)
        self.remember_point_focus = kwargs.get('remember_point_focus', None)
        self.skip_null_points = kwargs.get('skip_null_points', None)

    @property
    def mode(self) -> Optional[str]:
        """Sets the keyboard navigation mode for the chart. Defaults to ``'normal'``.

        Can be ``"normal"`` or ``"serialize"``.

        In ``'normal'`` mode, left/right arrow keys move between points in a series, while
        up/down arrow keys move between series. Up/down navigation acts intelligently to
        figure out which series makes sense to move to from any given point.

        In ``'serialize'`` mode, points are instead navigated as a single list.
        Left/right behaves as in ``"normal"`` mode. Up/down arrow keys will behave like
        left/right. This can be useful for unifying navigation behavior with/without
        screen readers enabled.

        :returns: The keyboard navigation mode to apply to the chart.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._mode

    @mode.setter
    def mode(self, value):
        if not value:
            self._mode = None
        else:
            value = validators.string(value).lower()
            if value not in ['normal', 'serialize']:
                raise errors.HighchartsValueError(f'Expects either "normal" or '
                                                  f'"serialize". Received: {value}')

            self._mode = value

    @property
    def point_navigation_enabled_threshold(self) -> Optional[bool | int]:
        """When a series contains more points than this, we no longer allow keyboard
        navigation for it. Accepts either an :class:`int <python:int>` or a
        a :class:`bool <python:bool>` value of ``False``.

        Setting to ``False`` will disable any threshold.

        :rtype: :class:`int <python:int>` or :class:`bool <python:bool>` or
          :obj:`None <python:None>`
        """
        return self._point_navigation_enabled_threshold

    @point_navigation_enabled_threshold.setter
    def point_navigation_enabled_threshold(self, value):
        if value is None:
            self._point_navigation_enabled_threshold = None
        elif isinstance(value, bool) and value is False:
            self._point_navigation_enabled_threshold = False
        else:
            value = validators.integer(value,
                                       allow_empty = False,
                                       coerce_value = True)
            self._point_navigation_enabled_threshold = value

    @property
    def remember_point_focus(self) -> Optional[bool]:
        """If ``True``, remembers which point was focused even after navigating away from
        the series, so that when navigating back to the series you start at the last
        focused point.

        Defaults to ``False``.

        :returns: Flag indicating whether to remember point focus.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._remember_point_focus

    @remember_point_focus.setter
    def remember_point_focus(self, value):
        if value is None:
            self._remember_point_focus = None
        else:
            self._remember_point_focus = bool(value)

    @property
    def skip_null_points(self) -> Optional[bool]:
        """If ``True``, skip null points when navigating through points with the keyboard.
        Defaults to ``True``.

        :returns: Flag indicating whether to skip null points in keyboard navigation.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._skip_null_points

    @skip_null_points.setter
    def skip_null_points(self, value):
        if value is None:
            self._skip_null_points = None
        else:
            self._skip_null_points = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'mode': as_dict.get('mode', None),
            'point_navigation_enabled_threshold': as_dict.get('pointNavigationEnabledThreshold',
                                                              None),
            'remember_point_focus': as_dict.get('rememberPointFocus', None),
            'skip_null_points': as_dict.get('skipNullPoints', None),
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'mode': self.mode,
            'pointNavigationEnabledThreshold': self.point_navigation_enabled_threshold,
            'rememberPointFocus': self.remember_point_focus,
            'skipNullPoints': self.skip_null_points
        }
        return untrimmed



# -*- coding: utf-8 -*-

"""
************************
Testing Philosophy
************************

.. note::

  Unit tests for **Highcharts for Python** are written using `pytest`_ and
  a comprehensive set of test automation are provided by `tox`_.

There are many schools of thought when it comes to test design. When building
**Highcharts for Python**, we decided to focus on practicality. That means:

  * **DRY is good, KISS is better.** To avoid repetition, our test suite makes
    extensive use of fixtures, parametrization, and decorator-driven behavior.
    This minimizes the number of test functions that are nearly-identical.
    However, there are certain elements of code that are repeated in almost all test
    functions, as doing so will make future readability and maintenance of the
    test suite easier.
  * **Coverage matters...kind of.** We have documented the primary intended
    behavior of every function in the **Highcharts for Python** library, and the
    most-likely failure modes that can be expected.  At the time of writing, we
    have about 94% code coverage. Yes, yes: We know that is less than 100%. But
    there are edge cases which are almost impossible to bring about, based on
    confluences of factors in the wide world. Our goal is to test the key
    functionality, and as bugs are uncovered to add to the test functions as
    necessary.

************************
Test Organization
************************

Each individual test module (e.g. ``test_chart.py``) corresponds to a
conceptual grouping of functionality. For example:

* ``tests/test_chart.py`` tests the contents of the :mod:`highcharts_core.chart` module.
* ``tests/options/annotations/test_annotations.py`` tests the contents of
  ``highcharts_core.options/annotations/__init__.py``.

Test Design
=====================

Because **Highcharts for Python** largely depends on designing classes that serialize
and deserialize content to/from JavaScript, we have implemented a standard number of
tests as Pytest fixtures. These tests are then configured using a set of parameters which
configure their standard inputs and expected outputs, as applicable.

If you are implementing a new class, please use the same test structure as used for the
other test modules. If you are adding a new method that is specific to one class, you do
not need to use this pattern. However, if it is a method that will be inherited by or used
by other classes, then we ask that you implement a similar pattern to keep test
maintenance as easy as possible.

**************************************
Configuring & Running Tests
**************************************

Installing with the Test Suite
=================================

.. tabs::

  .. tab:: Installing via pip

    .. code-block:: bash

      $ pip install highcharts-core[tests]

  .. tab:: From Local Development Environment

    .. seealso::

      When you
      :ref:`create a local development environment <preparing-development-environment>`,
      all dependencies for running and extending the test suite are installed.

Command-line Options
=====================

**Highcharts for Python** does not use any custom command-line options in its
test suite.

.. tip::

  For a full list of the CLI options, including the defaults available, try:

  .. code-block:: bash

    highcharts-core $ cd tests/
    highcharts-core/tests/ $ pytest --help

Configuration File
===================

Because **Highcharts for Python** has a very simple test suite, we have not
prepared a ``pytest.ini`` configuration file.

Running Tests
==============

.. tabs::

  .. tab:: Entire Test Suite

    .. code-block:: bash

      tests/ $ pytest

  .. tab:: Test Module

    .. code-block:: bash

      tests/ $ pytest tests/test_module.py

  .. tab:: Test Function

    .. code-block:: bash

      tests/ $ pytest tests/test_module.py -k 'test_my_test_function'

*****************
Skipping Tests
*****************

.. note::

  Because of the simplicity of **Highcharts for Python**, the test suite does
  not currently support any test skipping.

*******************
Incremental Tests
*******************

.. note::

  The **Highcharts for Python** test suite does support incremental testing using,
  however at the moment none of the tests designed rely on this functionality.

A variety of test functions are designed to test related functionality. As a
result, they are designed to execute incrementally. In order to execute tests
incrementally, they need to be defined as methods within a class that you decorate
with the ``@pytest.mark.incremental`` decorator as shown below::

    @pytest.mark.incremental
    class TestIncremental(object):
        def test_function1(self):
            pass
        def test_modification(self):
            assert 0
        def test_modification2(self):
            pass

This class will execute the ``TestIncremental.test_function1()`` test, execute and
fail on the ``TestIncremental.test_modification()`` test, and automatically fail
``TestIncremental.test_modification2()`` because of the ``.test_modification()``
failure.

To pass state between incremental tests, add a ``state`` argument to their method
definitions. For example::

    @pytest.mark.incremental
    class TestIncremental(object):
        def test_function(self, state):
            state.is_logged_in = True
            assert state.is_logged_in = True
        def test_modification1(self, state):
            assert state.is_logged_in is True
            state.is_logged_in = False
            assert state.is_logged_in is False
        def test_modification2(self, state):
            assert state.is_logged_in is True

Given the example above, the third test (``test_modification2``) will fail because
``test_modification`` updated the value of ``state.is_logged_in``.

.. note::

  ``state`` is instantiated at the level of the entire test session (one run of
  the test suite). As a result, it can be affected by tests in other test modules.

.. target-notes::

.. _`pytest`: https://docs.pytest.org/en/latest/
.. _`tox`: https://tox.readthedocs.io
.. _`mocks`: https://en.wikipedia.org/wiki/Mock_object
.. _`stubs`: https://en.wikipedia.org/wiki/Test_stub
"""



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
sys.path.insert(0, os.path.join(os.path.dirname(__name__), '..'))


version_dict = {}
with open(os.path.join(os.path.dirname(__file__),
                       '../',
                       'highcharts_core',
                       '__version__.py')) as version_file:
    exec(version_file.read(), version_dict)                   # pylint: disable=W0122

__version__ = version_dict.get('__version__')


project = 'Highcharts Core for Python'
copyright = '2022, HCP LLC'
author = 'Chris Modzelewski'

# The short X.Y version
version = __version__[:3]
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.graphviz',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx_tabs.tabs',
    'sphinx_toolbox.shields',
    'sphinx_toolbox.decorators',
    'sphinx_toolbox.issues',
    'sphinx_toolbox.formatting',
    'sphinx_toolbox.collapse',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Autodoc configuration settings.
autoclass_content = 'class'
autodoc_member_order = 'groupwise'
add_module_names = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

html_theme_options = {
    'navigation_depth': 3,
    'display_version': True,
    'prev_next_buttons_location': 'both',
    'style_external_links': False,
    'style_nav_header_background': 'rgb(70, 70, 92)'
}

html_css_files = [
    'sphinx_rtd_theme_ext_color_contrast.css'
]

html_js_files = [
    'https://code.jquery.com/jquery-3.6.4.min.js'
]

html_logo = '_static/highcharts-for-python-light-150x149.png'
html_favicon = '_static/highcharts-for-python-dark-32x32.png'

html_context = {
    "display_github": True,                 # Integrate GitHub
    "github_user": "highcharts-for-python",       # Username
    "github_repo": "highcharts-core",     # Repo name
    "github_version": "master",             # Version
    "conf_py_path": "/docs/",               # Path in the checkout to the docs root
}

github_username = 'highcharts-for-python'
github_repository = 'highcharts-core'

sphinx_tabs_disable_tab_closing = True

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3.10', None),
    'validator-collection': ('http://validator-collection.readthedocs.io/en/latest/', None),
    'ipython': ('https://ipython.readthedocs.io/en/stable/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'pyspark': ('https://spark.apache.org/docs/latest/api/python/', None),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Inheritance Diagram configuration
inheritance_graph_attrs = {
    'rankdir': 'TB'
}

suppress_warnings = [
    # 'ref.term',
    'ref.ref',
    'toc.not_readable'
]



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class SeriesAccessibility(HighchartsMeta):
    """Accessibility options global to all data series.

    .. hint::

      Individual series can also have specific accessibility options set.

    """

    def __init__(self, **kwargs):
        self._describe_single_series = None
        self._description_format = None
        self._description_formatter = None
        self._point_description_enabled_threshold = None

        self.describe_single_series = kwargs.get('describe_single_series', None)
        self.description_format = kwargs.get('description_format', None)
        self.description_formatter = kwargs.get('description_formatter', None)
        self.point_description_enabled_threshold = kwargs.get(
            'point_description_enabled_threshold',
            None
        )

    @property
    def describe_single_series(self) -> Optional[bool]:
        """If ``True``, will add series descriptions to charts with a single series.
        Defaults to ``False``.

        :returns: Flag indicating whether to add series descriptions to charts with a
          single series.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._describe_single_series

    @describe_single_series.setter
    def describe_single_series(self, value):
        if value is None:
            self._describe_single_series = None
        else:
            self._describe_single_series = bool(value)

    @property
    def description_format(self) -> Optional[str]:
        """Format to use for describing the data series group to assistive technology -
        including screen readers.

        Defaults to ``'{seriesDescription}{authorDescription}{axisDescription}'``.

        The series context and its subproperties are available under the variable
        ``{{series}}``, for example ``{{series.name}}`` for the series name, and
        ``{{series.points.length}}`` for the number of data points.

        The chart context and its subproperties are available under the variable
        ``{{chart}}``, for example ``{{chart.series.length}}`` for the number of series in
        the chart.

        ``{{seriesDescription}}`` refers to the automatic description of the series type
        and number of points added by Highcharts by default.

        ``{{authorDescription}}`` refers to the description added in
        ``series.description`` if one is present.

        ``{{axisDescription}}`` refers to the description added if the chart has multiple
        X or Y axes.

        Note that if :meth:`Series.description_formatter` is not :obj:`None <python:None>`
        it will take precedence, and this option will be overridden.

        :returns: Format string that applies to the description produced for the data
          series.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description_format

    @description_format.setter
    def description_format(self, value):
        self._description_format = validators.string(value, allow_empty = True)

    @property
    def description_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript formatter function to use instead of the default for series
        descriptions.

        Should receives one argument, series, referring to the series to describe. Should
        return a string with the description of the series for a screen reader user. If
        ``false`` is returned, the default formatter will be used for that series.

        :returns: JavaScript formatter function
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._description_formatter

    @description_formatter.setter
    @class_sensitive(CallbackFunction)
    def description_formatter(self, value):
        self._description_formatter = value

    @property
    def point_description_enabled_threshold(self) -> Optional[bool | int]:
        """When a series contains more points than the value set for this property,
        Highcharts will no longer expose information about individual points to screen
        readers.

        Defaults to ``200``.

        .. hint::

          If set to ``False``, the threshold will be disabled and all points will be
          described.

        :returns: The threshold for number of data points above which point description
          information wlil not be provided.
        :rtype: :class:`int <python:int>` or :class:`bool <python:bool>` or
          :obj:`None <python:None>`
        """
        return self._point_description_enabled_threshold

    @point_description_enabled_threshold.setter
    def point_description_enabled_threshold(self, value):
        if value is None:
            self._point_description_enabled_threshold = None
        elif isinstance(value, bool) and value is False:
            self._point_description_enabled_threshold = False
        else:
            self._point_description_enabled_threshold = validators.integer(
                value,
                allow_empty = False,
                coerce_value = True,
                minimum = 1
            )

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'describe_single_series': as_dict.get('describeSingleSeries', None),
            'description_format': as_dict.get('descriptionFormat', None),
            'description_formatter': as_dict.get('descriptionFormatter', None),
            'point_description_enabled_threshold': as_dict.get(
                'pointDescriptionEnabledThreshold',
                None
            ),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'describeSingleSeries': self.describe_single_series,
            'descriptionFormat': self.description_format,
            'descriptionFormatter': self.description_formatter,
            'pointDescriptionEnabledThreshold': self.point_description_enabled_threshold
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta


class LinkOptions(HighchartsMeta):
    """Link style options."""

    def __init__(self, **kwargs):
        self._color = None
        self._dash_style = None
        self._width = None

        self.color = kwargs.get('color', None)
        self.dash_style = kwargs.get('dash_style', None)
        self.width = kwargs.get('width', None)

    @property
    def color(self) -> Optional[str]:
        """Color of the link between two nodes. Defaults to ``rgba(100, 100, 100, 0.5)``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = validators.string(value, allow_empty = True)

    @property
    def dash_style(self) -> Optional[str]:
        """A name for the dash style to use for links. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dash_style

    @dash_style.setter
    def dash_style(self, value):
        self._dash_style = validators.string(value, allow_empty = True)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """Width (px) of the link between two nodes. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'color': as_dict.get('color', None),
            'dash_style': as_dict.get('dashStyle', None),
            'width': as_dict.get('width', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'color': self.color,
            'dashStyle': self.dash_style,
            'width': self.width
        }

        return untrimmed
